/*
 * Copyright (C) 2023 IOPSYS Software Solutions AB. All rights reserved.
 */
#if (EASYMESH_VERSION > 2)
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <ctype.h>

#include <sys/stat.h>
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>

#include <linux/if_bridge.h>
#include <netlink/socket.h>

#include <netinet/in.h>
#include <net/if.h>

#include <cmdu.h>
#include <1905_tlvs.h>
#include <easy/easy.h>
#include <easy/utils.h>
#include "wifi.h"

#include <i1905_wsc.h>

#include <uci.h>
#include <map_module.h>
#include <easymesh.h>
#include <dirent.h>

#include <openssl/err.h>
#include <openssl/bn.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/ec.h>


#include "timer.h"
#include "utils/1905_ubus.h"
#include "utils/utils.h"
#include "utils/debug.h"
#include "utils/dpp_sock.h"
#include "utils/keygen.h"
#include "utils/liblist.h"
#include "steer_rules.h"

#include "dpp.h"

#include "config.h"
#include "nl.h"
#include "agent.h"
#include "agent_ubus.h"
#include "agent_ubus_dbg.h"
#include "agent_map.h"
#include "cmdu_extension.h"
#include "agent_cmdu.h"
#include "agent_tlv.h"
#include "backhaul.h"

static void dpp_fd_timeout(atimer_t *t);

char *str2hex(char *cstr, int len, char *hexstr)
{
	int i, j;

	for (i = 0, j = 0; i < len; ++i, j += 2)
		sprintf(hexstr + j, "%02x", cstr[i] & 0xff);

	return hexstr;
}

char *agent_get_backhaul_hex(struct agent *a, enum wifi_band band,
			   char *encryption, char *hex_ssid, char *hex_key)
{
	struct netif_fh *fh;
	char *ifname = NULL;

	list_for_each_entry(fh, &a->fhlist, list) {
		char *ssid;
		char *key;

		if (!fh->cfg || !(fh->cfg->multi_ap & 0x01))
			continue;

		if (fh->cfg->band != band)
			continue;

		ssid = fh->cfg->ssid;
		key = fh->cfg->key;

		str2hex(ssid, strlen(ssid), hex_ssid);
		str2hex(key, strlen(key), hex_key);
		strncpy(encryption, fh->cfg->encryption, 32);
		ifname = (char *)fh->cfg->name;
		break;
	}

	return ifname;
}

uint8_t *dpp_get_attr(uint8_t *buf, size_t len, uint16_t req_id, uint16_t *ret_len)
{
	uint16_t id, alen;
	uint8_t *pos = buf, *end = buf + len;

	while (end - pos >= 4) {
		id = buf_get_le16(pos);
		pos += 2;
		alen = buf_get_le16(pos);
		pos += 2;
		if (alen > end - pos)
			return NULL;
		if (id == req_id) {
			*ret_len = alen;
			return pos;
		}
		pos += alen;
	}

	return NULL;
}


const char *dpp_type2str(enum dpp_bootstrap_type type)
{
	if (type == DPP_BOOTSTRAP_QR_CODE)
		return "qrcode";
	else if (type == DPP_BOOTSTRAP_PKEX)
		return "pkex";
	else if (type == DPP_BOOTSTRAP_NFC_URI)
		return "nfc";
	return "unknown";
}

const char *dpp_state2str(enum dpp_responder_state state)
{
	if (state == DPP_BOOSTRAPPING)
		return "bootstrapping";
	else if (state == DPP_AUTHENTICATING)
		return "authenticating";
	else if (state == DPP_AUTHENTICATED)
		return "authenticated";
	return "unknown";
}

const char *dpp_frame_type2str(uint8_t type)
{
	if (type == DPP_PA_AUTHENTICATION_REQ)
		return "Authentication Request";
	else if (type == DPP_PA_AUTHENTICATION_RESP)
		return "Authentication Response";
	else if (type == DPP_PA_AUTHENTICATION_CONF)
		return "Authentication Confirm";
	else if (type == DPP_PA_PEER_DISCOVERY_REQ)
		return "Peer Discovery Request";
	else if (type == DPP_PA_PEER_DISCOVERY_RESP)
		return "Peer Discovery Response";
	else if (type == DPP_PA_PKEX_V1_EXCHANGE_REQ)
		return "PKEX Version 1 Exchange Request";
	else if (type == DPP_PA_PKEX_EXCHANGE_RESP)
		return "PKEX Exchange Response";
	else if (type == DPP_PA_PKEX_COMMIT_REVEAL_REQ)
		return "PKEX Commit-Reveal Request";
	else if (type == DPP_PA_PKEX_COMMIT_REVEAL_RESP)
		return "PKEX Commit-Reveal Response";
	else if (type == DPP_PA_CONFIGURATION_RESULT)
		return "Configuration Result";
	else if (type == DPP_PA_CONNECTION_STATUS_RESULT)
		return "Connection Status Result";
	else if (type == DPP_PA_PRESENCE_ANNOUNCEMENT)
		return "Presence Announcement";
	else if (type == DPP_PA_RECONFIG_ANNOUNCEMENT)
		return "Reconfiguration Announcement";
	else if (type == DPP_PA_RECONFIG_AUTH_REQ)
		return "Reconfiguration Authentication Request";
	else if (type == DPP_PA_RECONFIG_AUTH_RESP)
		return "Reconfiguration Authentication Response";
	else if (type == DPP_PA_RECONFIG_AUTH_CONF)
		return "Reconfiguration Authentication Confirm";
	else if (type == DPP_PA_PKEX_EXCHANGE_REQ)
		return "PKEX Exchange Request";
	else if (type == 0x13)
		return "Push Button Presence Announcement";
	else if (type == 0x14)
		return "Push Button Presence Announcement Response";
	else if (type == 0x15)
		return "Private Peer Introduction Query";
	else if (type == 0x16)
		return "Private Peer Introduction Notify";
	else if (type == 0x17)
		return "Private Peer Introduction Update";
	return "Unknown";
}

uint8_t dpp_get_pub_af_type(uint8_t *frame, uint16_t framelen)
{
	if (framelen < 7)
		return 255;

	return *(frame + 6);
}

uint8_t dpp_get_gas_frame_type(const uint8_t *frame, uint16_t framelen)
{
	if (framelen < 1)
		return 255;

	return *frame;
}


struct dpp_enrollee *dpp_enrollee_get(struct agent *a, uint8_t *mac)
{
	struct dpp_enrollee *e;

	list_for_each_entry(e, &a->dpp_ctx.enrolleelist, list) {
		if (!memcmp(e->mac, mac, 6)) {
			return e;
		}
	}

	return NULL;
}

#if 0
struct dpp_enrollee *dpp_enrollee_get_hash(struct agent *a, uint8_t *hash)
{
	struct dpp_enrollee *e;

	list_for_each_entry(e, &a->dpp_ctx.enrolleelist, list) {
		if (e->hash && !memcmp(e->hash, hash, 32)) {
			return e;
		}
	}

	return NULL;
}
#endif

/* PROXY AGENT MODE FUNC */
struct dpp_enrollee *dpp_enrollee_alloc(struct agent *a, uint8_t *mac)
{
	struct dpp_enrollee *e;

	e = calloc(1, sizeof(*e));
	if (!e) {
		err("|%s:%d| Out of memory!\n", __func__, __LINE__);
		return NULL;
	}

	e->a = a;
	memcpy(e->mac, mac, 6);
	INIT_LIST_HEAD(&e->framelist);
	list_add(&e->list, &a->dpp_ctx.enrolleelist);
	return e;
}

void dpp_enrollee_free(struct dpp_enrollee *e)
{
	struct dpp_frame *f;

	list_for_each_entry(f, &e->framelist, list) {
		list_del(&f->list);
		free(f->data);
		free(f);
	}

	list_del(&e->list);
	free(e);
}

struct dpp_chirp *dpp_chirp_get(struct agent *a, uint8_t *hash)
{
	struct dpp_chirp *c;

	list_for_each_entry(c, &a->dpp_ctx.chirplist, list) {
		if (!memcmp(c->hash, hash, 32)) {
			return c;
		}
	}

	return NULL;
}

/* PROXY AGENT MODE FUNC */
struct dpp_chirp *dpp_chirp_alloc(struct agent *a, uint8_t *hash)
{
	struct dpp_chirp *c;

	c = calloc(1, sizeof(*c));
	if (!c) {
		err("|%s:%d| Out of memory!\n", __func__, __LINE__);
		return NULL;
	}

	c->a = a;
	memcpy(c->hash, hash, 32);
	list_add(&c->list, &a->dpp_ctx.chirplist);
	timer_init(&c->timeout, dpp_fd_timeout);

	return c;
}


/* ENROLLEE MODE FUNC */
int dpp_gen_uri(struct netif_bk *bk)
{
	struct agent *a = bk->agent;
	struct dpp_context *dpp = &a->dpp_ctx;
	struct dpp_uri_cfg *p = bk->dpp_cfg;
	char buf[256] = {0};
	char macstr[13] = {0};
	char chanstr[128] = {0}; /* at most (3+1+3+1)*16 - 1 = 127 characters */
	char *key;
	int i;

	if (!btostr(bk->bssid, 6, macstr)) {
		dbg("|%s:%d| mac byte array is corrupt!\n",
			__func__, __LINE__);
		return -1;
	}

	for (i = 0; i < p->num_chan; i++) {
		struct dpp_chan *chan = &p->chan[i];

		snprintf(chanstr + strlen(chanstr),
				sizeof(chanstr) - strlen(chanstr),
				"%u/%u,", chan->opclass, chan->channel);
	}

	/* remove excess ',' if any */
	if (strlen(chanstr) && chanstr[strlen(chanstr) - 1] == ',')
		chanstr[strlen(chanstr) - 1] = '\0';

	key = calloc(1, (dpp->ec_len * 2) + 1);
	if (!key)
		return -1;

	btostr(dpp->ec_key, dpp->ec_len, key);

	Cmd(buf, sizeof(buf), "/lib/wifi/multiap dpp gen_uri %s %s %s %s %s", p->ifname, dpp_type2str(p->type), macstr, chanstr, key);
	warn("|%s:%d| /lib/wifi/multiap dpp gen_uri %s %s %s %s", __func__, __LINE__, p->ifname, dpp_type2str(p->type), macstr, chanstr);
	if (!strlen(buf))
		return -1;

	strncpy(p->uri, buf, sizeof(p->uri));
	warn("|%s:%d| ifname:%s added uri:%s\n", __func__, __LINE__, p->ifname, p->uri);
	/* TODO: write to UCI */

	return 0;
}

/* ENROLLEE MODE FUNC */
int dpp_del_uri(struct netif_bk *bk)
{
	struct dpp_uri_cfg *p = bk->dpp_cfg;


	runCmd("/lib/wifi/multiap dpp del_uri %s", p->ifname);
	warn("|%s:%d| /lib/wifi/multiap dpp del_uri %s", __func__, __LINE__, p->ifname);

	memset(p->uri, 0, sizeof(p->uri));
	/* TODO: delete from UCI */

	return 0;
}

/* ENROLLEE MODE FUNC */
int dpp_chirp(struct netif_bk *bk)
{
	struct dpp_uri_cfg *p = bk->dpp_cfg;

	runCmd("/lib/wifi/multiap dpp chirp %s", p->ifname);
	warn("|%s:%d| /lib/wifi/multiap dpp chirp %s", __func__, __LINE__, p->ifname);

	return 0;
}

int dpp_listen_enrollee(struct netif_bk *bk)
{
	struct dpp_uri_cfg *p = bk->dpp_cfg;
	char freqstr[128] = {0}; /* at most (4+1)*16 = 80 characters  */
	int i;


	for (i = 0; i < p->num_chan; i++) {
		struct dpp_chan *chan = &p->chan[i];
		uint16_t freq = 0;

		if (p->band == BAND_2)
			freq = 2412 + (5 * (chan->channel - 1));
		else if (p->band == BAND_5)
			freq = 5170 + (5 * (chan->channel - 1));
		else {
			warn("|%s:%d| TODO: band:%d is not supported! (yet)\n",
				__func__, __LINE__, p->band);
			continue;
		}

		snprintf(freqstr + strlen(freqstr),
				sizeof(freqstr) - strlen(freqstr),
				"%d ", freq);
	}

	runCmd("/lib/wifi/multiap dpp listen %s %s", p->ifname, freqstr);
	warn("|%s:%d| /lib/wifi/multiap dpp listen %s %s", __func__, __LINE__, p->ifname, freqstr);

	return 0;
}

/* TODO: obvious gap 6GHz not handled! */
enum wifi_band wifi_channel_to_band(uint8_t channel)
{
        if (channel >= 1 && channel < 14)
                return BAND_2;
        else if (channel >= 36 && channel < 200)
                return BAND_5;

        return BAND_UNKNOWN;
}

static int ieee80211_chan_to_freq(uint8_t op_class, uint8_t chan)
{
	/* Table E-4 in IEEE Std 802.11-2012 - Global operating classes */
	switch (op_class) {
	case 81:
		/* channels 1..13 */
		if (chan < 1 || chan > 13)
			return -1;
		return 2407 + 5 * chan;
	case 82:
		/* channel 14 */
		if (chan != 14)
			return -1;
		return 2414 + 5 * chan;
	case 83: /* channels 1..9; 40 MHz */
	case 84: /* channels 5..13; 40 MHz */
		if (chan < 1 || chan > 13)
			return -1;
		return 2407 + 5 * chan;
	case 115: /* channels 36,40,44,48; indoor only */
	case 116: /* channels 36,44; 40 MHz; indoor only */
	case 117: /* channels 40,48; 40 MHz; indoor only */
	case 118: /* channels 52,56,60,64; dfs */
	case 119: /* channels 52,60; 40 MHz; dfs */
	case 120: /* channels 56,64; 40 MHz; dfs */
		if (chan < 36 || chan > 64)
			return -1;
		return 5000 + 5 * chan;
	case 121: /* channels 100-140 */
	case 122: /* channels 100-142; 40 MHz */
	case 123: /* channels 104-136; 40 MHz */
		if (chan < 100 || chan > 140)
			return -1;
		return 5000 + 5 * chan;
	case 124: /* channels 149,153,157,161 */
		if (chan < 149 || chan > 161)
			return -1;
		return 5000 + 5 * chan;
	case 125: /* channels 149,153,157,161,165,169,173,177 */
	case 126: /* channels 149,157,165,173; 40 MHz */
	case 127: /* channels 153,161,169,177; 40 MHz */
		if (chan < 149 || chan > 177)
			return -1;
		return 5000 + 5 * chan;
	case 128: /* center freqs 42, 58, 106, 122, 138, 155, 171; 80 MHz */
	case 130: /* center freqs 42, 58, 106, 122, 138, 155, 171; 80 MHz */
		if (chan < 36 || chan > 177)
			return -1;
		return 5000 + 5 * chan;
	case 129: /* center freqs 50, 114, 163; 160 MHz */
		if (chan < 36 || chan > 177)
			return -1;
		return 5000 + 5 * chan;
	case 131: /* UHB channels, 20 MHz: 1, 5, 9.. */
	case 132: /* UHB channels, 40 MHz: 3, 11, 19.. */
	case 133: /* UHB channels, 80 MHz: 7, 23, 39.. */
	case 134: /* UHB channels, 160 MHz: 15, 47, 79.. */
	case 135: /* UHB channels, 80+80 MHz: 7, 23, 39.. */
		if (chan < 1 || chan > 233)
			return -1;
		return 5950 + chan * 5;
	case 136: /* UHB channels, 20 MHz: 2 */
		if (chan == 2)
			return 5935;
		return -1;
	case 180: /* 60 GHz band, channels 1..8 */
		if (chan < 1 || chan > 8)
			return -1;
		return 56160 + 2160 * chan;
	case 181: /* 60 GHz band, EDMG CB2, channels 9..15 */
		if (chan < 9 || chan > 15)
			return -1;
		return 56160 + 2160 * (chan - 8);
	case 182: /* 60 GHz band, EDMG CB3, channels 17..22 */
		if (chan < 17 || chan > 22)
			return -1;
		return 56160 + 2160 * (chan - 16);
	case 183: /* 60 GHz band, EDMG CB4, channel 25..29 */
		if (chan < 25 || chan > 29)
			return -1;
		return 56160 + 2160 * (chan - 24);
	}
	return -1;
}


int dpp_parse_uri_chan_list(struct agent *a, struct dpp_enrollee *e, char *uri)
{
	char *pos, *pos2;
	int opclass = -1, channel, freq;

	e->num_chan = 0;

	dbg("|%s:%d| uri:%s\n", __func__, __LINE__, uri);

	pos = strstr(uri, "C:");
	if (!pos) {
		err("DPP: Invalid URI channel-list");
		return -1;
	}

	pos += 2; /* step past C: */

	while (pos && *pos && *pos != ';') {
		struct dpp_chan *chan;

		if (e->num_chan >= DPP_MAX_CH_OPCLASS)
			break;

		chan = &e->chan[e->num_chan];

		pos2 = pos;
		while (*pos2 >= '0' && *pos2 <= '9')
			pos2++;
		if (*pos2 == '/') {
			opclass = atoi(pos);
			pos = pos2 + 1;
		}
		if (opclass <= 0)
			goto fail;
		channel = atoi(pos);
		if (channel <= 0)
			goto fail;
		while (*pos >= '0' && *pos <= '9')
			pos++;

		freq = ieee80211_chan_to_freq(opclass, channel);
		dbg("DPP: URI channel-list: opclass=%d channel=%d ==> freq=%d",
			   opclass, channel, freq);

		chan->opclass = opclass;
		chan->channel = channel;
		chan->freq = freq;
		e->num_chan++;

		if (*pos == ';' || *pos == '\0')
			break;
		if (*pos != ',')
			goto fail;
		pos++;
	}

fail:
	return 0;
}

bool dpp_band_has_uri(struct agent *a, enum wifi_band band)
{
	struct dpp_enrollee *e;

	list_for_each_entry(e, &a->dpp_ctx.enrolleelist, list) {
		if (e->band == band)
			return true;
	}

	return false;
}

/* PROXY AGENT MODE FUNC */
void dpp_sta_event_handler(void *agent, struct blob_attr *msg)
{
	struct agent *a = (struct agent *)agent;
	char ifname[16] = {0}, event[16] = {0};
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	struct netif_fh *fh;

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	fh = get_netif_by_name(a, ifname);
	if (!fh)
		return;

	if (fh->cfg->multi_ap == 0x01)
		return;

	if (!strcmp(event, "action")) {
		struct blob_attr *data[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "raw", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0};
		uint8_t enrollee[6] = {0};
		char *framestr;
		uint8_t *frame;
		int framelen = 0;

		blobmsg_parse(data_attr, 2, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1])
			return;

		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);

		if (!hwaddr_aton(mac_str, enrollee))
			return;

		framestr = (char *)blobmsg_data(data[1]);

		framelen = strlen(framestr);
		framelen = framelen / 2;
		frame = calloc(framelen, sizeof(uint8_t));
		if (!frame)
			return;

		if (!strtob(framestr, framelen, frame)) {
			free(frame);
			return;
		}

		if (*frame == 0x04) {
			struct dpp_enrollee *e;
			uint8_t *ptr = frame;
			struct dpp_frame *dppf;

			e = dpp_enrollee_get(a, enrollee);
			if (!e) {
				dbg("|%s:%d| no dpp enrollee for mac "MACFMT", allocate\n",
				__func__, __LINE__, MAC2STR(enrollee));
				e = dpp_enrollee_alloc(a, enrollee);
				if (!e) {
					dbg("|%s:%d| Failed to allocate enrollee for mac "MACFMT"\n",
					__func__, __LINE__, MAC2STR(enrollee));
					goto out;
				}
			}

			ptr++;

			if (FRAME_IS_DPP_PUB_AF(ptr)) {
				struct cmdu_buff *cmdu;
				int frametype;

				ptr += 6; /* PA Type + oui + oui type + crypto suite */

				warn("|%s:%d| received DPP action frame!\n",
					__func__, __LINE__);


				frametype = dpp_get_pub_af_type(frame + 1, framelen - 1);
				if (frametype == 255)
					goto out;

				dppf = dpp_enrollee_get_frame(a, e, frametype);
				if (dppf && !timestamp_expired(&dppf->time, 500) &&
				    framelen == dppf->len && !memcmp(dppf->data, frame, framelen)) {
					warn("|%s:%d| identical frame handled withing last 500ms from enrolle:"MACFMT" dropping frame type:%d\n",
					     __func__, __LINE__, MAC2STR(enrollee), frametype);
					goto out;
				}

				ptr++; /* frame type */
				if (frametype == DPP_PA_PRESENCE_ANNOUNCEMENT) {
					uint16_t attrlen = 0;
					uint8_t *pos;

					pos = dpp_get_attr(ptr,
							framelen - labs(ptr - frame),
							DPP_ATTR_R_BOOTSTRAP_KEY_HASH,
							&attrlen);
					if (pos) {
						struct dpp_chirp *c;

						c = dpp_chirp_get(a, pos);
						if (!c) {
							dbg("|%s:%d| matching chrip not found, allocate\n",
							    __func__, __LINE__);
							c = dpp_chirp_alloc(a, pos);
							if (!c) {
								dbg("|%s:%d| Failed to allocate chirp\n",
								    __func__, __LINE__);
								goto out;
							}
						}

						e->chirp = c;
						if (c->auth_req && c->chirp_fd.fd > 0) {
							struct dpp_frame *auth_req = c->auth_req;

							socket_send_dpp_tcp_message(c->chirp_fd.fd,
								auth_req->data,
								auth_req->len);
							warn("|%s:%d| already have assoc req frame!\n", __func__, __LINE__);
							goto out;
						}

					}
				}

				ptr += 5;

				if (e->chirp) {
					cmdu = agent_gen_proxied_encap_dpp(a,
							enrollee,
							frametype,
							framelen - 1, frame + 1,
							32,
							e->chirp->hash);

					if (cmdu) {
						agent_send_cmdu(a, cmdu);
						cmdu_free(cmdu);
					}
				}
				dpp_enrollee_add_frame(a, e, frametype, framelen, frame);
			} else if (FRAME_IS_DPP_GAS_FRAME(ptr)) {
				struct cmdu_buff *cmdu;
				int frametype;


				ptr++;

				frametype = dpp_get_gas_frame_type(frame + 1, framelen - 1);
				if (frametype == 255)
					goto out;

				if (e->chirp) {
					cmdu = agent_gen_proxied_encap_dpp(a,
							enrollee,
							frametype,
							framelen - 1, frame + 1,
							32,
							e->chirp->hash);
					if (cmdu) {
						agent_send_cmdu(a, cmdu);
						cmdu_free(cmdu);
					}
				}

				dpp_enrollee_add_frame(a, e, frametype, framelen, frame);
			}
		}
out:
		free(frame);
	}
}


int dpp_enrollee_update_hash(struct agent *a,
			     struct dpp_enrollee *e,
			     uint8_t len, uint8_t *hash)
{
	struct dpp_chirp *c;

	if (e->chirp) {
		c = e->chirp;
		memcpy(c->hash, hash, 32);
		return 0;
	}

	c = dpp_chirp_get(a, hash);
	if (!c) {
		dbg("|%s:%d| no dpp chirp for hash, allocate\n",
			__func__, __LINE__);

		c = dpp_chirp_alloc(a, hash);
		if (!c) {
			dbg("|%s:%d| Failed to allocate enrollee for hash\n",
				__func__, __LINE__);
			return -1;
		}
	}

	e->chirp = c;
	return 0;
}


struct dpp_frame *dpp_enrollee_get_frame(struct agent *a,
					 struct dpp_enrollee *e, uint8_t type)
{
	struct dpp_frame *frame;

	list_for_each_entry(frame, &e->framelist, list) {
		if (type == frame->type)
			return frame;
	}

	return NULL;
}

struct dpp_frame *dpp_enrollee_add_frame(struct agent *a, struct dpp_enrollee *e,
			   uint8_t type, uint16_t len, uint8_t *data)
{
	struct dpp_frame *frame;

	frame = dpp_enrollee_get_frame(a, e, type);
	if (!frame) {
		frame = calloc(1, sizeof(*frame));
		if (!frame) {
			err("|%s:%d| Out of memory!\n", __func__, __LINE__);
			return NULL;
		}

		frame->type = type;
		list_add(&frame->list, &e->framelist);
	} else {
		free(frame->data);
		frame->data = NULL;
		frame->len = 0;
	}

	frame->data = calloc(1, len);
	if (!frame->data) {
		err("|%s:%d| Out of memory!\n", __func__, __LINE__);
		list_del(&frame->list);
		free(frame);
		return NULL;
	}

	timestamp_update(&frame->time);
	memcpy(frame->data, data, len);

	frame->len = len;
	return frame;
}

void dpp_configurator_evh(void *agent, struct blob_attr *msg)
{
	struct agent *a = (struct agent *) agent;
	static const struct blobmsg_policy ev_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_STRING },
	};
	struct blob_attr *tb[ARRAY_SIZE(ev_attr)];
	char *ifname, *event, *data;

	blobmsg_parse(ev_attr, ARRAY_SIZE(ev_attr), tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1] || !tb[2])
		return;

	ifname = blobmsg_data(tb[0]);
	event = blobmsg_data(tb[1]);
	data =  blobmsg_data(tb[2]);

	if (!strcmp(event, "dpp-connector")) {
		warn("|%s:%d| ifname:%s dpp_connector:%s\n", __func__, __LINE__, ifname, data);
		runCmd("hostapd_cli -i %s set dpp_connector \"%s\"", ifname, data);
	} else if (!strcmp(event, "dpp-c-sign-key")) {
		warn("|%s:%d| ifname:%s dpp-c-sign-key:%s\n", __func__, __LINE__, ifname, data);
		runCmd("hostapd_cli -i %s set dpp_csign \"%s\"", ifname, data);
	} else if (!strcmp(event, "dpp-net-access-key")) {
		warn("|%s:%d| ifname:%s dpp-net-access-key:%s\n", __func__, __LINE__, ifname, data);
		runCmd("hostapd_cli -i %s set dpp_netaccesskey \"%s\"", ifname, data);
	}

	UNUSED(a);

	return;
}

static unsigned char hexdigit2int(unsigned char xd)
{
	if (xd <= '9')
		return xd - '0';
	xd = tolower(xd);
	if (xd == 'a')
		return 10;
	if (xd == 'b')
		return 11;
	if (xd == 'c')
		return 12;
	if (xd == 'd')
		return 13;
	if (xd == 'e')
		return 14;
	if (xd == 'f')
		return 15;
  	return 0;
}

int hexstr2str(const char *hexstr, int buflen, char *buf)
{
	const char *src = hexstr;
	char *dst = buf;
	int i = 0;

	while (*src != '\0' && buflen > i) {
		const unsigned char high = hexdigit2int(*src++);
		const unsigned char low  = hexdigit2int(*src++);
		*dst++ = (high << 4) | low;
		i++;
	}

	*dst = '\0';
	return 0;
}

void dpp_enrollee_evh(void *agent, struct blob_attr *msg)
{
	struct agent *a = (struct agent *) agent;
	static const struct blobmsg_policy ev_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_STRING },
	};
	struct blob_attr *tb[ARRAY_SIZE(ev_attr)];
	char *ifname, *event, *data;
	struct netif_bk *bk;
	struct dpp_creds *creds;

	blobmsg_parse(ev_attr, ARRAY_SIZE(ev_attr), tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1] || !tb[2])
		return;

	ifname = blobmsg_data(tb[0]);
	event = blobmsg_data(tb[1]);
	data = blobmsg_data(tb[2]);

	bk = agent_get_netif_bk_by_name(a, ifname);
	if (!bk)
		return;

	creds = &bk->creds;

	if (!strcmp(event, "dpp-confobj-akm")) {
		char *akm = creds->akm;

		warn("|%s:%d| dpp-confobj-akm:%s\n", __func__, __LINE__, data);

		strncpy(akm, data, sizeof(creds->akm));
		creds->stored |= DPP_CONFOBJ_AKM;
	} else if (!strcmp(event, "dpp-confobj-ssid")) {
		char *ssid = creds->ssid;

		warn("|%s:%d| dpp-confobj-ssid:%s\n", __func__, __LINE__, data);

		strncpy(ssid, data, sizeof(creds->ssid));
		creds->stored |= DPP_CONFOBJ_SSID;
	} else if (!strcmp(event, "dpp-confobj-pass")) {
		char *key;
		int len = 0;

		len = (strlen(data) / 2) + 1;

		key = calloc(1, len);
		if (!key)
			return;

		hexstr2str(data, len, key);
		warn("|%s:%d| dpp-confobj-pass:%s\n", __func__, __LINE__, data);
		warn("|%s:%d| dpp-confobj-pass-ascii:%s\n", __func__, __LINE__, key);

		strncpy(creds->key, key, sizeof(creds->key));
		creds->stored |= DPP_CONFOBJ_KEY;
	} else if (!strcmp(event, "dpp-connector")) {
		int len = strlen(data);

		warn("|%s:%d| dpp-connector:%s\n", __func__, __LINE__, data);
		if (!len)
			return;

		if (creds->connector)
			free(creds->connector);

		creds->connector = calloc(1, len + 1);
		if (!creds->connector)
			return;

		strncpy(creds->connector, data, len);
		creds->stored |= DPP_CONNECTOR;
		timer_set(&bk->dpp_do_chirp, 60 * 1000);
	} else if (!strcmp(event, "dpp-c-sign-key")) {
		int len = strlen(data);

		warn("|%s:%d| dpp-c-sign-key:%s\n", __func__, __LINE__, data);

		if (!len)
			return;

		if (creds->csign)
			free(creds->csign);

		creds->csign = calloc(1, len + 1);
		if (!creds->csign)
			return;

		strncpy(creds->csign, data, len);
		creds->stored |= DPP_CSIGN;
	} else if (!strcmp(event, "dpp-net-access-key")) {
		int len = strlen(data);

		warn("|%s:%d| dpp-net-access-key:%s\n", __func__, __LINE__, data);

		if (!len)
			return;

		if (creds->netaccesskey)
			free(creds->netaccesskey);

		creds->netaccesskey = calloc(1, len + 1);
		if (!creds->netaccesskey)
			return;

		strncpy(creds->netaccesskey, data, len);
		creds->stored |= DPP_NETACCESSKEY;
	}

	if ((creds->stored & 0b0111) == 0b0111) {
		struct wifi_radio_element *re;
		struct agent_config_radio *rcfg;

		warn("|%s:%d| All credentials are received for ifname:%s! Trigger write\n", __func__, __LINE__, bk->name);
		wifi_apply_iface_cfg(bk->name, creds->akm,
				     creds->ssid, creds->key);
		timer_set(&a->onboarding_scheduler, 5 * 1000);
		creds->stored = creds->stored & ~(DPP_CONFOBJ_SSID | DPP_CONFOBJ_KEY | DPP_CONFOBJ_AKM);

		bk->cfg->onboarded = true;
		re = wifi_ifname_to_radio_element(a, bk->name);
		if (re)
			re->onboarded = 1;
		uci_set_wireless_interface_option("mapagent", "radio", "device",
						  bk->cfg->device, "onboarded",
						  "1");

		rcfg = get_agent_config_radio(&a->cfg, bk->cfg->device);
		if (rcfg)
			rcfg->onboarded = 1;
	} else if ((creds->stored & (DPP_NETACCESSKEY | DPP_CONNECTOR | DPP_CSIGN))) {
		struct wifi_radio_element *re;
		struct agent_config_radio *rcfg;

		warn("|%s:%d| DPP creds received for ifname:%s!\n", __func__, __LINE__, bk->name);

		//timer_set(&a->onboarding_scheduler, 5 * 1000);
		creds->stored = creds->stored & ~(DPP_NETACCESSKEY | DPP_CONNECTOR | DPP_CSIGN);

		bk->cfg->dpp_onboarded = true;
		re = wifi_ifname_to_radio_element(a, bk->name);
		if (re)
			re->dpp_onboarded = 1;
		uci_set_wireless_interface_option("mapagent", "radio", "device",
						  bk->cfg->device, "dpp_onboarded",
						  "1");

		uci_set_wireless_interface_option("mapagent", "bsta", "ifname",
						  bk->cfg->name, "dpp_onboarded",
						  "1");

		rcfg = get_agent_config_radio(&a->cfg, bk->cfg->device);
		if (rcfg)
			rcfg->dpp_onboarded = 1;
	}


	return;
}

int dpp_get_keys(struct agent *a)
{
	struct dpp_context *dpp = &a->dpp_ctx;
	struct blob_buf keys = { 0 };
	struct blob_attr *key;
	int rem;
	int rc = -1;
	static const struct blobmsg_policy attr[] = {
		[0] = { .name = "keys", .type = BLOBMSG_TYPE_ARRAY },
	};
	struct blob_attr *tb[ARRAY_SIZE(attr)];

	blob_buf_init(&keys, 0);

	if (!blobmsg_add_json_from_file(&keys, DPP_KEY_FILE)) {
		warn("Failed to parse %s\n", DPP_KEY_FILE);
		goto out;
	}

	blobmsg_parse(attr, ARRAY_SIZE(attr), tb, blob_data(keys.head), blob_len(keys.head));
	if (!tb[0])
		goto out;

	blobmsg_for_each_attr(key, tb[0], rem) {
		static const struct blobmsg_policy uri_attr[4] = {
			[0] = { .name = "type", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "key", .type = BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *tb1[ARRAY_SIZE(uri_attr)];
		char *key_str;
		char *type;
		uint8_t *key_buf;
		int len;

		blobmsg_parse(uri_attr, ARRAY_SIZE(uri_attr), tb1, blobmsg_data(key), blob_len(key));
		if (!tb1[0] || !tb1[1])
			continue;

		type = blobmsg_data(tb1[0]);
		if (!type)
			continue;

		key_str = blobmsg_data(tb1[1]);
		if (!key_str)
			continue;

		len = strlen(key_str) / 2;

		key_buf = calloc(1, len);
		if (!key_buf)
			continue;

		strtob(key_str, len, key_buf);

		if (!strcmp(type, "ec_priv")) {
			dpp->ec_len = len;
			dpp->ec_key = key_buf;
		} else
			free(key_buf);
	}

out:
	blob_buf_free(&keys);
	return rc;
}

int dpp_append_key_file(struct agent *a, uint8_t *key, int key_len, char *type)
{
	struct json_object *key_json, *arr, *val, *new;
	int rv, len, i;
	char *key_str;

	key_str = calloc(1, (key_len * 2) + 1);
	if (!key_str)
		return -1;

	btostr(key, key_len, key_str);

	key_json = json_object_from_file(DPP_KEY_FILE);
	if (!key_json) {
		warn("|%s:%d| failed to read json:%s error:%s. "\
		     "Try to generate new\n", __func__, __LINE__, DPP_KEY_FILE,
		     json_util_get_last_err());

		key_json = json_object_new_object();
		if (!key_json)
			goto out;

		arr = json_object_new_array();
		if (!arr)
			goto out_uri;

		json_object_object_add(key_json, "keys", arr);
	}

	rv = json_object_object_get_ex(key_json, "keys", &arr);
	if (!rv || !arr)
		goto out_uri;

	if (!json_object_is_type(arr, json_type_array)) {
		warn("|%s:%d| file:%s is not expected format\n", __func__,
		     __LINE__, DPP_KEY_FILE);
		goto out_uri;
	}

	len = json_object_array_length(key_json);
	for (i = 0; i < len; i++) {
		struct json_object *t;
		const char *p;

		t = json_object_array_get_idx(arr, i);
		if (!t) {
			warn("|%s:%d| couldn't get entry:%d from uri array",
			     __func__, __LINE__, i);
			continue;
		}

		p = json_get_string(t, "type");
		if (strncmp(p, type, 32)) /* arbitrary length */
			continue;

		/* hash already exists in file */
		/* TODO: possibly type has changed? */
		goto out;
	}

	new = json_object_new_object();
	if (!new)
		goto out_uri;

	val = json_object_new_string(key_str);
	if (!val) {
		json_object_put(new);
		goto out_uri;
	}

	json_object_object_add(new, "key", val);

	val = json_object_new_string(type);
	if (!val) {
		json_object_put(new);
		goto out_uri;
	}

	json_object_object_add(new, "type", val);
	json_object_array_add(arr, new);
	json_object_to_file(DPP_KEY_FILE, key_json);
out_uri:
	json_object_put(key_json);
out:
	free(key_str);
	return 0;
}

int dpp_gen_keypair(struct agent *a)
{
	struct dpp_context *dpp = &a->dpp_ctx;
	int rc;

	rc = ec_gen_keypair(&dpp->ec_key, &dpp->ec_len);
	if (rc < 0)
		return rc;

	dpp_append_key_file(a, dpp->ec_key, dpp->ec_len, "ec_priv");

	return 0;
}

uint16_t dpp_get_controller_tcp_port(struct agent *a, enum wifi_band band)
{
	struct dpp_controller_cfg *dpp_cntlr;

	dpp_cntlr = get_dpp_controller_by_band(&a->cfg, band);
	if (dpp_cntlr) {
		return dpp_cntlr->port;
	} else {
		if (band == BAND_2)
			return 8902;
		else if (band == BAND_5)
			return 8905;
	}

	return 8910;
}

static void dpp_fd_timeout(atimer_t *t)
{
	struct dpp_chirp *c = container_of(t, struct dpp_chirp, timeout);

	socket_close(&c->chirp_fd.fd);
	uloop_fd_delete(&c->chirp_fd);
}

void dpp_period_chirp(atimer_t *t)
{
	struct netif_bk *bk = container_of(t, struct netif_bk, dpp_do_chirp);
	struct agent *a = bk->agent;

	if (!bk->dpp_cfg) {
		dbg("|%s:%d| no chirp cfg found for bsta:%s\n", __func__,
		    __LINE__, bk->name);
		timer_set(&bk->dpp_do_chirp, 10 * 1000);
		return;
	}

	if ((!bk->cfg->onboarded && !bk->cfg->dpp_onboarded) /* backhaul has not been onboarded */ ||
	    (!agent_has_active_backhaul(a) && timestamp_expired(&a->disconnect_t, 60)) /* we have had no uplink for over 60s */ ||
	    (!agent_has_active_backhaul(a) && bk->cfg->dpp_onboarded && !bk->cfg->onboarded ) /* DPP was used for uplink and connection is lost */ ) {
		/* TODO: URI gen should probably be elsewhere OR this should be agent wide, not per bsta */
		dpp_gen_uri(bk);
		dpp_chirp(bk);
		dpp_listen_enrollee(bk);
	}

	timer_set(&bk->dpp_do_chirp, bk->dpp_cfg->chirp_int * 1000);
}

void dpp_client_event_uloop_cb(struct uloop_fd *fd, unsigned int events)
{
	struct dpp_chirp *c = container_of(fd, struct dpp_chirp, chirp_fd);
	uint8_t *test;
	int len;

	dbg("|%s:%d| Receiving enrollee tcp message from fd:%d!\n", __func__,
	    __LINE__, c->chirp_fd.fd);
	len = dpp_recv_tcp_data(c->chirp_fd.fd, &test);
	if (len < 0) {
		dbg("Read error! Close socket:%d\n", c->chirp_fd.fd);
		socket_close(&c->chirp_fd.fd);
		uloop_fd_delete(&c->chirp_fd);
		timer_del(&c->timeout);
		return;
	} else if (len == 0) {
		dbg("No data to be read on socket:%d\n", c->chirp_fd.fd);
		return;
	}

	dump(test, len, "DPP Frame from hostapd");

	return;
}


void dpp_event_uloop_cb(struct uloop_fd *fd, unsigned int events)
{
	struct dpp_context *dpp = container_of(fd, struct dpp_context, dpp_sk);
	struct agent *a = container_of(dpp, struct agent, dpp_ctx);
	int client_fd;
	uint8_t *test;
	int len;
	uint8_t frametype;

	dbg("|%s:%d| Receiving new tcp message (chirp?) from fd:%d!\n",
	    __func__, __LINE__, dpp->dpp_sk.fd);

	client_fd = socket_dpp_tcp_server_rx(dpp->dpp_sk.fd);
	if (client_fd < 0)
		return;

	len = dpp_recv_tcp_data(client_fd, &test);
	if (len <= 0) {
		dbg("No data was read!\n");
		return;
	}

	dump(test, len, "DPP Frame from hostapd");

	frametype = dpp_get_pub_af_type(test, len);
	if (frametype == 255) {
		dbg("Error getting frame type!\n");
		return;
	}

	/* This socket SHOULD not receive any other frames -- can probably
	 * return if frame type does not match
	 */
	if (frametype == DPP_PA_PRESENCE_ANNOUNCEMENT) {
		uint16_t attrlen = 0;
		uint8_t *pos;
		struct dpp_chirp *c;
		int ret;

		pos = dpp_get_attr(test + 7, len - 7,
				   DPP_ATTR_R_BOOTSTRAP_KEY_HASH,
				   &attrlen);
		if (!pos) {
			dbg("did not find hash, exit!\n");
			return;
		}

		dump(pos, 32, "hash");

		c = dpp_chirp_get(a, pos);
		if (!c) {
			dbg("|%s:%d| no dpp chirp for hash, allocate\n",
			    __func__, __LINE__);

			c = dpp_chirp_alloc(a, pos);
			if (!c) {
				dbg("|%s:%d| Failed to allocate enrollee for hash\n",
				    __func__, __LINE__);
				return;
			}
		}

		if (c->chirp_fd.fd <= 0) {
			c->chirp_fd.cb = dpp_client_event_uloop_cb;
			c->chirp_fd.fd = client_fd;
			ret = uloop_fd_add(&c->chirp_fd, ULOOP_READ);
			timer_set(&c->timeout, 20 * 1000); /* 20s */
			dbg("|%s:%d| setup dpp_sk -- ret:%d fd:%d\n", __func__,
			    __LINE__, ret, c->chirp_fd.fd);
		} else {
			dbg("|%s:%d| chirp already has fd!! fd:%d\n", __func__,
			    __LINE__, c->chirp_fd.fd);
		}
	}

	free(test);
}
#endif
